/*
  dbimodule.c
  $Revision: 1.6 $
  $Date: 1997/02/19 13:40:06 $

  This is the general interface to COPPERMAN compliant database
  interface.

  In particular, types and type numbers are defined
  
  dbi.c
  Date: 1997/05/15 11:00:00 
  Jayant Kelkar (jayant@mindspring.com)
  Added support for DATE/DATETIME/INTERVAL types.
  Now they behave as expected. Created a python class IfxDTObject
  which encapsulates the details and lives in DbiDate.
 */

#include <time.h>

#include "allobjects.h"
#include "intobject.h"


#define const
#define DBI_EXPORT

#include "dbi.h"

/* create a IfxDTObject */
static PyObject *IfxDateTimeError; /* Exception */

staticforward PyTypeObject IfxDateTime_Type;
typedef struct
{
  PyObject_HEAD
  dtime_t dateTime; /* datetime if DATETIME */
  int     date;     /* date if DATE */
  int     type;     /* DATE = 0, DATETIME = 1, INTERVAL = 2 */
  short   null;     /* 1 of null 0 otherwise */
} IfxDTObject;


/* IfxDTObject member access */
#define dtDateTime(v) (((IfxDTObject *) (v))->dateTime)
#define dtDate(v)     (((IfxDTObject *) (v))->date)
#define dtType(v)     (((IfxDTObject *) (v))->type)
#define dtNull(v)     (((IfxDTObject *) (v))->null)

static PyObject *
ifxDateFromDT(PyObject *self)
{
  long date =  (dtDate(self));
  char tmp[11];
  int rc ;
  int sz = 10;
  memset(tmp, 0, 11);
  rc = rdatestr(date, tmp);
  if (tmp[0] == ' ')
    sz = 0;
  return (PyString_FromStringAndSize(tmp, sz));
}

long
ifxLongFromDT(PyObject *self)
{
  IfxDTObject *v = (IfxDTObject*) self;
  long l = (dtDate(self));
  return l;
}

static PyObject *
ifxDateTimeFromDT(PyObject *self)
{
  dtime_t datetime =  (dtDateTime(self));
  char tmp[27];
  int rc ;
  memset(tmp, 0, 27);
  rc = dttofmtasc(&datetime, tmp, 26, NULL /*"yyyy-mm-dd hh:mm:ss.sss"*/);
  if (rc != 0)
    printf("in ifxDateTimeFromDT: rc = %d\n", rc);
  
  return (PyString_FromString(tmp));
}

dtime_t
ifxdtime_tFromDT(PyObject *self)
{
  IfxDTObject *v = (IfxDTObject*) self;
  dtime_t datetime = (dtDateTime(self));
  return datetime;
}

static IfxDTObject *
New_IfxDTObjectFromDate(PyObject *self, PyObject *args)
{
  /* create an object from a date */
  IfxDTObject *v;
  long longdate;
  int rc;
  char *format, *input;
  
  if (!getargs(args, "(ss)", &format, &input)) 
    return NULL;
  rc = rdefmtdate(&longdate, format, input);
  switch (rc) {
  case 0:
    break;
  case -1204:
    printf("input contains an invalid year component\n");
    return NULL;
  case -1205:
    printf("input contains an invalid month component\n");
    return NULL;
  case -1206:
    printf("input contains an invalid day component\n");
    return NULL;
  case -1209:
    printf("The input contains no delimiters, length should be only 6\n");
    return NULL;
      case -1212:
    printf("The input does not contain all of year, month and day components\n");
    return NULL;
  default:
    printf("Unexpected return code %d\n", rc);
  }
  
  v = PyObject_NEW(IfxDTObject, &IfxDateTime_Type);
  dtDate(v) = (int)longdate;
  dtType(v) = DATE;
  return v;
}

static IfxDTObject *
New_IfxDTObjectFromDateTime(PyObject *self, PyObject *args)
{
  IfxDTObject *v;
  dtime_t datetime;
  int rc;
  char *format, *input;
  
  if (!getargs(args, "(ss)", &format, &input)) 
    return NULL;
  rc = dtcvfmtasc(input, format, datetime);
  if (rc < 0) {
    printf("conversion failed\n");
    return NULL;
  }
      
  v = PyObject_NEW(IfxDTObject, &IfxDateTime_Type);
  dtDateTime(v) = datetime;
  dtType(v) = DATETIME;
  return v;
}

static IfxDTObject * 
newIfxdtFromDate(long date, short *null)
{
  short type = 0; /* date */
  /* create a new object from a date */
  IfxDTObject *v;
  v = PyObject_NEW(IfxDTObject, &IfxDateTime_Type);
  
  if (v == NULL) {
    fprintf(stderr, "Failed to malloc memory for IfxDTObject\n");
    return NULL;
  }
  if (*null == -1)  /* this data is null */
    dtNull(v) = 1; 
  else
    dtNull(v) = 0;
  
  if (*null == -1) 
    printf("Date is NULL\n");
  dtDate(v) = date;
  dtType(v) = DATE;          /* as date */
  return v;
}
    
static IfxDTObject * 
newIfxdtFromDateTime(dtime_t *data, short *null, short type)
{
  /* create a new object from a datetime */
  IfxDTObject *v;
  
  v = PyObject_NEW(IfxDTObject, &IfxDateTime_Type);
  if (v == NULL) {
    fprintf(stderr, "Failed to malloc memory for IfxDTObject\n");
    return NULL;
  }
  {
    char tmp[25];
    tmp[24] = '\0';
    memcpy(&dtDateTime(v), data, 24);
    dtType(v) = type ; /* type is set in ifxdbmoduel.ec */
    if (*null == -1)  /* this data is null */
      dtNull(v) = 1;
    else
      dtNull(v) = 0;
  }
  return v;
}



static PyMethodDef ifxdtMethods[] = {
  /*{"ifxdtfd",  New_IfxDTObjectFromDate,     1}, 
    {"ifxdtfdt", New_IfxDTObjectFromDateTime, 1},  */
  {NULL,       NULL}  /* sentinel */
};

static PyObject *
dateStr(PyObject *self)
{
  PyObject *v = (PyObject *)dbiValue(self);
  if (dtType(v) == DATE)
    /* this is a date */
    return ifxDateFromDT(v);
  else {
    return ifxDateTimeFromDT(v);
  }
}

static void
ifxdtDealloc(PyObject *obj)
{
  DEL(obj);
}

static PyObject *
ifxdtGetAttr(PyObject *self, char *name)
{
  return Py_FindMethod(ifxdtMethods, self, name);
}

static int
ifxdtPrint(PyObject *self, FILE *fp, int flags)
{
  if (dtType(self) == DATE) {
    char date[11];
    date[10] = '\0';
    rdatestr(ifxLongFromDT(self), date);
    if (date[0] == ' ')
      date[0] = '\0';
    else 
      fprintf(fp, "%s", date);
    return 0;
  }

  if (dtType(self) == DATETIME ||
      dtType(self) == INTERVAL) {
    char datetime[27];
    /* print the contents - based on the type */
    dtime_t *d = &dtDateTime(self);
    memset(datetime, 0, 27);
    dttofmtasc(d, datetime, 26, NULL); /* use the default format string */
    fprintf(fp, "%s", datetime);
    return 0;
  }
}

static PyObject *
ifxdtRepr(PyObject *self)
{
  if (dtType(self) == DATE) {
    char date[15];
    date[14] = '\0';
    strncpy(date, "DATE ", 5);
    rdatestr(ifxLongFromDT(self), &date[5]);
    if (date[5] == ' ')
      date[5] = '\0';
    return PyString_FromString(date);
  }

  if (dtType(self) == DATETIME ||
      dtType(self) == INTERVAL) {
    char datetime[36];
    /* print the contents - based on the type */
    dtime_t *d = &dtDateTime(self);
    if (dtType(self) == DATETIME)
      strncpy(datetime, "DATETIME ", 9);
    if (dtType(self) == INTERVAL)
      strncpy(datetime, "INTERVAL ", 9);
    memset(&datetime[9], 0, 27);
    dttofmtasc(d, &datetime[9], 26, NULL); /* use the default format string */
    return PyString_FromString(datetime);
  }
  
}


static PyTypeObject IfxDateTime_Type =
{
#ifdef WIN32
  PyObject_HEAD_INIT (0)
#else
  PyObject_HEAD_INIT (&PyType_Type)
#endif
  0,			/*ob_size */
  "IfxDateTime",	/*tp_name */
  sizeof(IfxDTObject),	/*tp_basicsize */
  0,			/*tp_itemsize */
  ifxdtDealloc,	        /*tp_dealloc */
  ifxdtPrint,		/*tp_print */
  ifxdtGetAttr,		/*tp_getattr */
  0,                    /*tp_setattr */
  0,                    /*tp_compare */
  0,                    /*tp_repr */
  0, /*&dt_as_number,   / **tp_as_number */
  0,                    /**tp_as_sequence */
  0,                    /**tp_as_mapping */
  0,                    /*tp_hash */
  0,                    /*tp_call */
  0                     /*tp_str */
} ;

typedef struct
{
  PyObject_HEAD
  PyObject *objectOf;
} DbiContainer;


PyObject *
dbiValue(PyObject *o)
{
  return ((DbiContainer *) o)->objectOf;
}

static PyObject *makeDbiTypeTP(PyObject *args, PyTypeObject *ty)
{
  DbiContainer *dbt = 0;
  PyObject *ob;
  
  if (PyArg_ParseTuple(args, "O", &ob)) {
    dbt = PyObject_NEW(DbiContainer, ty);
    dbt->objectOf = ob;
	Py_INCREF(ob);
  }
  return (PyObject*)dbt;
}

static void 
dbiDealloc(PyObject *self)
{
  Py_DECREF(((DbiContainer *) self)->objectOf);
  PyMem_DEL(self);
}

static PyMethodDef noMethods[] = {
  0, 0
};

static PyObject *
dbiGetAttr(PyObject *self,
	   char *name)
{
  if (!strcmp(name, "value")) {
    PyObject *val = dbiValue(self);
    Py_INCREF(val);
    return val;
  }
  return Py_FindMethod(noMethods, self, name);
}

static int
dtPrint(PyObject *self, FILE *fp, int flags)
{
  PyObject *v = (PyObject*)dbiValue(self);
  return ifxdtPrint(v, fp, flags);
}

static PyObject *
dtRepr(PyObject *self)
{
  PyObject *v = (PyObject*)dbiValue(self);
  return ifxdtRepr(v);
}

#define delg(a) dbiValue(a)->ob_type->tp_as_number

static PyObject* 
dt_nb_add(PyObject* a, PyObject* b)
{ return delg(a)->nb_add(dbiValue(a),b); }

static PyObject* 
dt_nb_subtract(PyObject* a, PyObject* b)
{ return delg(a)->nb_subtract(dbiValue(a),b); }

static PyObject* 
dt_nb_int(PyObject* a)
{ return delg(a)->nb_int(dbiValue(a)); }

static PyObject* 
dt_nb_long(PyObject* a)
{ return delg(a)->nb_long(dbiValue(a)); }

static PyObject* 
dt_nb_float(PyObject* a)
{ return delg(a)->nb_float(dbiValue(a)); }

static PyObject* 
dt_nb_oct(PyObject* a)
{ return delg(a)->nb_oct(dbiValue(a)); }

static PyObject* 
dt_nb_hex(PyObject* a)
{ return delg(a)->nb_hex(dbiValue(a)); }

static int 
dt_cmp(PyObject *a, PyObject *b)
{ return dbiValue(a)->ob_type->tp_compare(a,b); }

static int 
dt_nb_coerce(PyObject **, PyObject **);

PyNumberMethods dt_as_number = {
  dt_nb_add ,   /* nb_add */
  dt_nb_subtract ,   /* nb_subtract */
  0,   		/* nb_multiply */
  0,   		/* nb_divide */
  0,   		/* nb_remainder */
  0,   		/* nb_divmod */
  0,   		/* nb_power */
  0,   		/* nb_negative */
  0,   		/* nb_positive */
  0,   		/* nb_absolute */
  0,   		/* nb_nonzero */
  0,   		/* nb_invert */
  0,   		/* nb_lshift */
  0,   		/* nb_rshift */
  0,   		/* nb_and */
  0,   		/* nb_xor */
  0,   		/* nb_or */
  dt_nb_coerce, /* nb_coerce */
  dt_nb_int ,   /* nb_int */
  dt_nb_long ,  /* nb_long */
  dt_nb_float , /* nb_float */
  dt_nb_oct ,   /* nb_oct */
  dt_nb_hex     /* nb_hex */
} ;

static PyTypeObject DbiDate_Type =
{
#ifdef WIN32
  PyObject_HEAD_INIT (0)
#else
  PyObject_HEAD_INIT (&PyType_Type)
#endif
  0,			/*ob_size */
  "DbiDate",		/*tp_name */
  sizeof(DbiContainer),	/*tp_basicsize */
  0,			/*tp_itemsize */
  dbiDealloc,		/*tp_dealloc */
  dtPrint,		/*tp_print */
  dbiGetAttr,		/*tp_getattr */
  0,                    /*tp_setattr */
  dt_cmp,               /*tp_compare */
  dtRepr,     		/*tp_repr */
  &dt_as_number,        /**tp_as_number */
  0,                    /**tp_as_sequence */
  0,                    /**tp_as_mapping */
  0,                    /*tp_hash */
  0,                    /*tp_call */
  dateStr               /*tp_str */
} ;


static int 
dt_nb_coerce(PyObject **pv, PyObject **pw)
{
  if (PyNumber_Check(*pw)) {
    DbiContainer *dbt = PyObject_NEW(DbiContainer, &DbiDate_Type);
    dbt->objectOf = PyNumber_Int(*pw);
    Py_INCREF(*pv);
    *pw = (PyObject *)dbt;
    return 0;
  }
  return 1; /* Can't do it */
}

static PyTypeObject DbiRaw_Type =
{
#ifdef WIN32
  PyObject_HEAD_INIT (0)
#else
  PyObject_HEAD_INIT (&PyType_Type)
#endif
  0,			/*ob_size */
  "DbiRaw",		/*tp_name */
  sizeof(DbiContainer),	/*tp_basicsize */
  0,			/*tp_itemsize */
  dbiDealloc,   	/*tp_dealloc */
  0,			/*tp_print */
  dbiGetAttr,		/*tp_getattr */
  0,                    /*tp_setattr */
  0,                    /*tp_compare */
  0,                    /*tp_repr */
  0,                    /**tp_as_number */
  0,                    /**tp_as_sequence */
  0,                    /**tp_as_mapping */
  0,                    /*tp_hash */
  0,                    /*tp_call */
  dbiValue              /*tp_str */
} ;

static PyTypeObject DbiRowId_Type =
{
#ifdef WIN32
  PyObject_HEAD_INIT (0)
#else
  PyObject_HEAD_INIT (&PyType_Type)
#endif
  0,			/*ob_size */
  "DbiRowId",		/*tp_name */
  sizeof(DbiContainer),	/*tp_basicsize */
  0,			/*tp_itemsize */
  dbiDealloc,   	/*tp_dealloc */
  0,			/*tp_print */
  dbiGetAttr		/*tp_getattr */
} ;

static PyObject *
makeDate(PyObject *self, PyObject *args)
{
  return makeDbiTypeTP(args, &DbiDate_Type);
}

static PyObject *
makeRaw(PyObject *self, PyObject *args)
{
  return makeDbiTypeTP(args, &DbiRaw_Type);
}

static PyObject *
makeRowId(PyObject *self, PyObject *args)
{
  return makeDbiTypeTP(args, &DbiRowId_Type);
}

static PyObject *
makeDbiType(PyObject *o, PyTypeObject *ty)
{
  DbiContainer *dbt = PyObject_NEW(DbiContainer, ty);
  dbt->objectOf = o;
  return (PyObject*) dbt;
}

PyObject *
dbiMakeDate(PyObject *contents)
{
  return makeDbiType(contents, &DbiDate_Type);
}

PyObject *
dbiMakeRaw(PyObject *contents)
{
  return makeDbiType(contents, &DbiRaw_Type);
}

PyObject *
dbiMakeRowId(PyObject *contents)
{
  return makeDbiType(contents, &DbiRowId_Type);
}

/* these are here because statics are only visible in one file */
PyObject *
ifxMakeDateFromDate(long date, short *null)
{
  return (PyObject *)newIfxdtFromDate(date, null);
}

PyObject *
ifxMakeDateFromDateTime(dtime_t *data, short *null, short type)
{
  return (PyObject *)newIfxdtFromDateTime(data, null, type);
}

static PyMethodDef globalMethods[] = {
  { "dbDate",  makeDate, 1} ,
  { "dbiDate", makeDate, 1} ,
  { "dbRaw",   makeRaw, 1} ,
  { "dbiRaw",  makeRaw, 1} ,
  {0,          0}        /* Sentinel */
};

void 
initdbi()
{
#ifdef WIN32
  PyObject *m, *d;
  DbiDate_Type.ob_type = &PyType_Type;	
  DbiRaw_Type.ob_type = &PyType_Type;	
  DbiRowId_Type.ob_type = &PyType_Type;	
  m = Py_InitModule("dbi", globalMethods);
  d = PyModule_GetDict(m);
#else
  PyObject *m = Py_InitModule("dbi", globalMethods);
  PyObject *d = PyModule_GetDict(m);
#endif
  PyDict_SetItemString(d, "STRING",
		       DbiString = PyString_FromString("STRING"));
  PyDict_SetItemString(d, "RAW",
		       DbiRaw = PyString_FromString("RAW"));
  PyDict_SetItemString(d, "NUMBER",
		       DbiNumber = PyString_FromString("NUMBER"));
  PyDict_SetItemString(d, "DATE",
		       DbiDate = PyString_FromString("DATE"));
  PyDict_SetItemString(d, "ROWID",
		       DbiRowId = PyString_FromString("ROWID"));
  PyDict_SetItemString(d, "TYPES",
		       Py_BuildValue("(OOOOO)",
				     DbiString,
				     DbiRaw,
				     DbiNumber,
				     DbiDate,
				     DbiRowId));

  /* Establish errors */
  PyDict_SetItemString(d, "noError",
		       DbiNoError = PyString_FromString("dbi.no-error"));
  PyDict_SetItemString(d, "opError",
		       DbiOpError = PyString_FromString("dbi.operation-error"));
  PyDict_SetItemString(d, "progError",
		       DbiProgError = PyString_FromString("dbi.program-error"));
  PyDict_SetItemString(d, "integrityError",
		       DbiIntegrityError =
		       PyString_FromString("dbi.integrity-error"));
  PyDict_SetItemString(d, "dataError",
		       DbiDataError = PyString_FromString("dbi.data-error"));
  PyDict_SetItemString(d, "internalError",
		       DbiInternalError = 
		       PyString_FromString("dbi.internal-error"));

}

int dbiIsDate(const PyObject *o)
{
  return o->ob_type == &DbiDate_Type;
}

int dbiIsRaw(const PyObject *o)
{
  return o->ob_type == &DbiRaw_Type;
}

int dbiIsRowId(const PyObject *o)
{
  return o->ob_type == &DbiRowId_Type;
}
